package com.ichinaski.imageloader;

import java.io.FileDescriptor;
import java.lang.reflect.Field;
import java.util.zip.CRC32;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.util.DisplayMetrics;
import android.view.Display;
import android.view.ViewGroup.LayoutParams;
import android.view.WindowManager;
import android.view.animation.AlphaAnimation;
import android.view.animation.Animation;
import android.view.animation.DecelerateInterpolator;
import android.widget.ImageView;

import static com.ichinaski.imageloader.ImageLoader.TAG;
import static com.ichinaski.imageloader.util.LogUtils.LOGD;
import static com.ichinaski.imageloader.util.LogUtils.LOGE;

public class ImageLoadUtils {
    
    public static int calculateInSampleSize(ImageSize size, ImageSize reqSize) {
        if (reqSize == null) {
            return 1;
        }
        
        final int width = size.width;
        final int height = size.height;
        final int reqWidth = reqSize.width;
        final int reqHeight = reqSize.height;
        
        int inSampleSize = 1;

        if (height > reqHeight || width > reqWidth) {
            if (width > height) {
                inSampleSize = Math.round((float)height / (float)reqHeight);
            } else {
                inSampleSize = Math.round((float)width / (float)reqWidth);
            }
            
            // This offers some additional logic in case the image has a strange
            // aspect ratio. For example, a panorama may have a much larger
            // width than height. In these cases the total pixels might still
            // end up being too large to fit comfortably in memory, so we should
            // be more aggressive with sample down the image (=larger
            // inSampleSize).
            final float totalPixels = width * height;
            
            // Anything more than 2x the requested pixels we'll sample down further.
            final float totalReqPixelsCap = reqWidth * reqHeight * 2;
            
            while (totalPixels / (inSampleSize * inSampleSize) > totalReqPixelsCap) {
                inSampleSize++;
            }
        }
        return inSampleSize;
    }

    /**
     * Hash image URL to be suitable for using as a disk filename.
     */
    public static String getImageId(String _url) {
        CRC32 crc32 = new CRC32();
        crc32.update(_url.getBytes());
        return String.valueOf(crc32.getValue());
    }
    
    /**
     * Decode and sample down a bitmap from an input stream to the requested ImageSize.
     *
     * @param fileDescriptor The file descriptor to read from
     * @param size The requested ImageSize of the resulting bitmap
     * @return A bitmap sampled down from the original with the same aspect ratio and dimensions
     *         that are equal to or greater than the requested width and height
     */
    public static BitmapSample decodeSampledBitmap(FileDescriptor fileDescriptor, ImageSize size) {
        return decodeSampledBitmap(null, fileDescriptor, size);
    }
    
    /**
     * Decode and sample down a bitmap from a file to the requested ImageSize.
     *
     * @param filename The full path of the file to decode
     * @param size The requested ImageSize of the resulting bitmap
     * @return A bitmap sampled down from the original with the same aspect ratio and dimensions
     *         that are equal to or greater than the requested width and height
     */
    public static BitmapSample decodeSampledBitmap(String fileName, ImageSize size) {
        return decodeSampledBitmap(fileName, null, size);
    }
    
    private static BitmapSample decodeSampledBitmap(String fileName, 
            FileDescriptor descriptor, ImageSize size) {
        // Check dimensions
        BitmapFactory.Options options = new BitmapFactory.Options();
        options.inJustDecodeBounds = true;
        
        if (fileName != null) {
            BitmapFactory.decodeFile(fileName, options);
        } else {
            BitmapFactory.decodeFileDescriptor(descriptor, null, options);
        }
        
        // Calculate sample size
        ImageSize imgSize = new ImageSize(options.outWidth, options.outHeight);
        final int sampleSize = calculateInSampleSize(imgSize, size);
        options.inSampleSize = sampleSize;
        
        // Decode with inSampleSize set
        options.inJustDecodeBounds = false;
        Bitmap bmp = null;
        if (fileName != null) {
            bmp = BitmapFactory.decodeFile(fileName, options);
        } else {
            bmp = BitmapFactory.decodeFileDescriptor(descriptor, null, options);
        }
        
        if (bmp != null) {
            LOGD(TAG, "Image decoded. Original: " +imgSize.width + "x" + imgSize.height
                    + ". Sample: " + sampleSize 
                    + ". Sampled img: " + bmp.getWidth() + "x" + bmp.getHeight());
            return new BitmapSample(bmp, sampleSize, imgSize);
        }
        return null;
    }
    
    /**
     * Size computing algorithm:
     * 1) Get layout_width and layout_height. If both of them haven't exact value then go to step #2.
     * 2) Get maxWidth and maxHeight. If both of them are not set then go to step #3.
     * 3) Get device screen dimensions.
     */
    public static ImageSize getImageSize(ImageView imageView) {
        DisplayMetrics displayMetrics = imageView.getContext().getResources().getDisplayMetrics();

        LayoutParams params = imageView.getLayoutParams();
        int width = params.width; // Get layout width parameter
        if (width <= 0) width = getFieldValue(imageView, "mMaxWidth"); // Check maxWidth parameter
        if (width <= 0) width = displayMetrics.widthPixels;

        int height = params.height; // Get layout height parameter
        if (height <= 0) height = getFieldValue(imageView, "mMaxHeight"); // Check maxHeight parameter
        if (height <= 0) height = displayMetrics.heightPixels;

        return new ImageSize(width, height);
    }
    
    /**
     * Access the properties by Reflection.
     * 
     * @param object
     * @param fieldName
     * @return
     */
    private static int getFieldValue(Object object, String fieldName) {
        int value = 0;
        try {
            Field field = ImageView.class.getDeclaredField(fieldName);
            field.setAccessible(true);
            int fieldValue = (Integer) field.get(object);
            if (fieldValue > 0 && fieldValue < Integer.MAX_VALUE) {
                value = fieldValue;
            }
        } catch (Exception e) {
            LOGE(TAG, e.getMessage());
        }
        return value;
    }

    public static ImageSize getScreenSize(Context context) {
        DisplayMetrics metrics = new DisplayMetrics();
        Display display = ((WindowManager)context.getSystemService(Context.WINDOW_SERVICE))
                .getDefaultDisplay();
        display.getMetrics(metrics);

        return new ImageSize(metrics.widthPixels, metrics.heightPixels);
    }

    public static Animation fadeInAnimation() {
        Animation fadeInAnimation = new AlphaAnimation(0, 1);
        fadeInAnimation.setInterpolator(new DecelerateInterpolator());
        fadeInAnimation.setDuration(300);
        return fadeInAnimation;
    }

    public static int getScreenDensity(Context context) {
        DisplayMetrics metrics = new DisplayMetrics();
        Display display = ((WindowManager)context.getSystemService(Context.WINDOW_SERVICE))
                .getDefaultDisplay();
        display.getMetrics(metrics);
        return metrics.densityDpi;
    }

}
