package com.ichinaski.imageloader;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Collections;
import java.util.Map;
import java.util.Vector;
import java.util.WeakHashMap;

import android.app.Activity;
import android.content.Context;
import android.graphics.Bitmap;
import android.os.Process;
import android.text.TextUtils;
import android.view.View;
import android.widget.ImageView;

import com.ichinaski.imageloader.ImageListener.ImageAnimation;
import com.ichinaski.imageloader.cache.ImageCache;
import com.ichinaski.imageloader.cache.ImageCache.ImageCacheParams;
import com.ichinaski.imageloader.util.IOUtils;
import com.ichinaski.imageloader.util.PlatformUtils;

import static com.ichinaski.imageloader.util.LogUtils.LOGD;
import static com.ichinaski.imageloader.util.LogUtils.LOGE;
import static com.ichinaski.imageloader.util.LogUtils.LOGV;

public class ImageLoader {
    public static final String TAG = ImageLoader.class.getSimpleName();
    
    public interface ImageLoaderCallback {
        public void onImageLoaded(ImageView iv, Bitmap bmp, String url);
    }

    // Singleton
    private static ImageLoader mInstance;

    /**
     * NOTE: I'm using a Vector (which is a synchronised data structure). If at some point this
     * type is no longer a Vector, some methods of this class should be synchronised.
     */
    private Vector<ImageToDownload> mImagesQueue = new Vector<ImageToDownload>();
    private Map<View, String> mViews = Collections.synchronizedMap(new WeakHashMap<View, String>());

    private Context mContext;
    private ImageCache mImageCache;
    private ImageLoadConfig mConfig;
    private ImageDownloadThread[] mDownloadThreads;
    
    /**
     * Initializes ImageLoader instance with default values. MUST BE CALLED BEFORE CALLING getImageLoader().
     * Should be called in the application's onCreate() method, to ensure its instantiation
     * for later calls.
     * 
     * @param context
     * @param config
     */
    public static void init (Context context) {
        init(context, new ImageLoadConfig(context));
    }

    /**
     * Initializes ImageLoader instance. MUST BE CALLED BEFORE CALLING getImageLoader().
     * Should be called in the application's onCreate() method, to ensure its instantiation
     * for later calls.
     * 
     * @param context
     * @param config
     */
    public static void init (Context context, ImageLoadConfig config) {
        if (mInstance == null) {
            mInstance = new ImageLoader(context, config);
        }
    }
    
    /**
     * Instance getter. If init() has not been called previously, it will be called automatically.
     * 
     * @param context
     * @return ImageLoader instance
     */
    public static ImageLoader getImageLoader(Context context) {
        if (mInstance == null) {
          init(context, new ImageLoadConfig(context));
        }
        return mInstance;
    }

    private ImageLoader(Context context, ImageLoadConfig config) {
        // Retain the Application Context, to avoid memory leaks
        mContext = context.getApplicationContext();
        mConfig = config;

        mImageCache = new ImageCache(mContext, mConfig.cacheParams);

        mDownloadThreads = new ImageDownloadThread[mConfig.maxConnections];
        for (int i = 0; i < config.maxConnections; i++) {
            mDownloadThreads[i] = new ImageDownloadThread();
            mDownloadThreads[i].setName("ImageLoader " + i);// Util for logs
            mDownloadThreads[i].start();
        }
    }

    /**
     * Display the content of a certain URL into a given ImageView. If the ImageView is already
     * on the queue, its previous request will be cleaned, being replaced by this one.
     * The sample size will be deduced from the ImageView, and the default displaying options
     * will be applied. To use a custom listener, use 
     *   displayImage(String url, Activity activity, ImageView imageView, ImageSize size,
     *           IImageLoaderCallback onLoadListener)
     *           
     * @param url
     * @param activity
     * @param imageView
     */
    public void displayImage(String url, Activity activity, ImageView imageView) {
        displayImage(url, activity, imageView, null);
    }

    /**
     * Display the content of a certain URL into a given ImageView. If the ImageView is already
     * on the queue, its previous request will be cleaned, being replaced by this one.
     * The default displaying options will be applied. To use a custom listener, use 
     *   displayImage(String url, Activity activity, ImageView imageView, ImageSize size,
     *           IImageLoaderCallback onLoadListener)
     * 
     * @param url
     * @param activity
     * @param imageView
     * @param size
     */
    public void displayImage(String url, Activity activity, ImageView imageView, ImageSize size) {
        if (size == null) size = ImageLoadUtils.getImageSize(imageView);
        displayImage(url, activity, imageView, size, new ImageListener(ImageAnimation.NONE));
    }

    /**
     * Display the content of a certain URL into a given ImageView. If the ImageView is already
     * on the queue, its previous request will be cleaned, being replaced by this one.
     * 
     * @param url
     * @param activity
     * @param imageView
     * @param size
     * @param onLoadListener
     */
    public void displayImage(String url, Activity activity, ImageView imageView, ImageSize size,
            ImageLoaderCallback onLoadListener) {
        // Be nice
        if (TextUtils.isEmpty(url)) {
            imageView.setImageDrawable(null);
            return;
        }
        
        ImageToDownload image = new ImageToDownload(activity, imageView, url, size, onLoadListener);
        
        BitmapSample bitmapSample = mImageCache.getBitmapFromMemCache(url);
        if (bitmapSample != null) {
            final int requiredSampleSize = ImageLoadUtils.calculateInSampleSize(
                    bitmapSample.getOriginalSize(), size);
            final int cachedSampleSize = bitmapSample.getSampleSize();
            
            LOGD(TAG, "Bitmap found on memory cache. Cached sample size: " + cachedSampleSize 
                    + ". Required sample size: " + requiredSampleSize);
            
            // Well keep on memory cache last required sample size.
            // TODO: Keep track of this ImageToDownload's sample size computation and save it
            // to be used if has to be queued.
            if (requiredSampleSize == bitmapSample.getSampleSize()) {
                // Display
                LOGD(TAG, "Displaying...");
                display(image.imageView, bitmapSample.getBitmap(), image.url, image.listener);
            } else {
                LOGD(TAG, "Queuing...");
                //mImageCache.clearBitmap(url);
                queueImageToLoad(image);
            }
        } else {
            queueImageToLoad(image);
        }
    }

    private void queueImageToLoad(ImageToDownload image) {
        synchronized (mImagesQueue) {
            if (image.imageView != null) {
                cleanImageView(image.imageView);
            }
            mViews.put(image.imageView, image.url);
            mImagesQueue.add(image);
            mImagesQueue.notifyAll();
        }
    }

    /**
     * If the queue contains this image view's previous request, remove it.
     * 
     * @param imageView
     */
    public void cleanImageView(ImageView imageView) {
        // This method may be called from external users, thus the Reentrant Lock.
        synchronized (mImagesQueue) {
            for (ImageToDownload img : mImagesQueue) {
                if (imageView.equals(img.imageView)) {
                    LOGD(TAG, "Cleaning ImageView to be used by a new incoming Bitmap...");
                    mImagesQueue.remove(img);
                    break;
                }
            }
        }
    }

    // ============= Image Downloader ============= //
    class ImageDownloadThread extends Thread {
        public void run() {
            // Setup thread priority
            Process.setThreadPriority(Process.myTid(), Process.THREAD_PRIORITY_BACKGROUND);
            
            try {
                while (true) {
                    ImageToDownload image = null;

                    // Wait until we have images in the queue
                    synchronized (mImagesQueue) {
                        if (mImagesQueue.size() == 0) {
                            mImagesQueue.wait();
                        } else {
                            image = mImagesQueue.remove(0);
                        }
                    }

                    if (image != null) {
                        Bitmap bitmap = processImage(image);
                        dispatchImage(image, bitmap);
                    }

                    if (Thread.interrupted()) break;
                }
            }

            // Allow thread to exit
            catch (InterruptedException e) {}
        }
    }

    /**
     * 
     * @param image ImageToDownload
     * @return
     */
    private Bitmap processImage(ImageToDownload image) {
        BitmapSample bitmapSample = null;

        // from disk cache
        bitmapSample = mImageCache.getBitmapFromDiskCache(image.url, image.size);

        // from web
        if (bitmapSample == null) {
            final File tmpFile = downloadImage(image);
            if (tmpFile != null) {
                // Decode sampled image
                bitmapSample = ImageLoadUtils.decodeSampledBitmap(tmpFile.getAbsolutePath(), image.size);
                
                // Disk cache and delete tmp file
                if (bitmapSample != null) {
                    mImageCache.copyFileToDiskCache(image.url, tmpFile);
                }
                tmpFile.delete();
            }
        }
        
        // Cache resized Bitmap in memory cache.
        if (bitmapSample != null) {
            mImageCache.addBitmapToMemoryCache(image.url, bitmapSample);
        }

        return bitmapSample != null ? bitmapSample.getBitmap() : null;
    }
        
    /**
     * If the awaited url for this img view is the same as the given
     * ImageToDownload, or, if the tag is null (i.e. no associated ImageView),
     * we dispatch the image. Otherwise, do nothing. If there's no tag for
     * this image, the listener will handle the bitmap.
     * 
     * @param image Just processed image.
     * @param bitmap Downloaded bitmap.
     */
    private void dispatchImage(ImageToDownload image, Bitmap bitmap) {
        final String tag = mViews.get(image.imageView);
        if ((image.url.equals(tag)) || image.imageView == null) {
            BitmapDisplayer bd = new BitmapDisplayer(image.imageView, bitmap, image.url, image.listener);
            Activity a = image.activity;
            a.runOnUiThread(bd);
        } else if (tag != null) {
            LOGD(Thread.currentThread().getName(), 
                    "TAG IS DIFFERENT. Processed URL: " + image.url + ". Awaited URL: " + tag);
        } else {
            // Should never happen
            LOGE(Thread.currentThread().getName(), "WTF?!");
        }
    }

    /**
     * Download an image from a URL, write it to a disk and return the File
     * 
     * @param ImageToDownload
     * @return A File pointing to the fetched bitmap
     */
    private File downloadImage(ImageToDownload image) {
        final File cacheDir = PlatformUtils.getDiskCacheDir(mContext, mConfig.cacheParams.uniqueName);

        if (!cacheDir.exists()) {
            cacheDir.mkdir();
        }

        LOGV(TAG, "Downloading " + image.url);
        InputStream is = null;
        OutputStream os = null;
        File tmpFile = null;
        HttpURLConnection conn = null;

        try {
            final String url = image.url.replace(" ", "%20");
            tmpFile = File.createTempFile("bitmap", null, cacheDir);

            URL imageUrl = new URL(url);
            conn = (HttpURLConnection)imageUrl.openConnection();
            conn.setConnectTimeout(mConfig.CONN_TIMEOUT);
            conn.setReadTimeout(mConfig.READ_TIMEOUT);
            
            is = conn.getInputStream();
            os = new FileOutputStream(tmpFile);
            IOUtils.copyStream(is, os);
        } catch (IOException ex) {
            tmpFile = null;
            LOGE(TAG, ex.getMessage());
        } finally {
            if (conn != null) conn.disconnect();
            IOUtils.closeSilently(is);
            IOUtils.closeSilently(os);
        }
        return tmpFile;
    }

    class BitmapDisplayer implements Runnable {
        ImageLoaderCallback listener;
        ImageView imageView;
        Bitmap bitmap;
        String url;

        public BitmapDisplayer(ImageView iv, Bitmap bmp, String url, ImageLoaderCallback l) {
            this.listener = l;
            this.bitmap = bmp;
            this.imageView = iv;
            this.url = url;
        }

        @Override
        public void run() {
            display(imageView, bitmap, url, listener);
        }
    }

    /*
    // Provide a default listener to simply display the image
    private IImageLoaderCallback defaultListener = new IImageLoaderCallback() {
        @Override
        public void onImageLoaded(ImageView iv, Bitmap bmp, String url) {
            if (iv != null && bmp != null) {
                iv.setImageBitmap(bmp);
            }
        }
    };
    */
    
    private void display(ImageView imageView, Bitmap bitmap, String url, ImageLoaderCallback listener) {
        listener.onImageLoaded(imageView, bitmap, url);
    }

    public void clearCache() {
        clearMemoryCache();
        clearDiskCache();
    }
    
    public void clearMemoryCache() {
        mImageCache.clearMemoryCache();
    }
    
    public void clearDiskCache() {
        mImageCache.clearDiskCache();
    }
    
    public void finalize () {
        if (mImageCache != null) {
            mImageCache.close();
        }
    }

    public static class ImageLoadConfig {
        private static final int DEFAULT_CONN_TIMEOUT = 10000;
        private static final int DEFAULT_READ_TIMEOUT = 10000;
        private static final int DEFAULT_MAX_CONNECTIONS = 4;

        public int CONN_TIMEOUT = DEFAULT_CONN_TIMEOUT;
        public int READ_TIMEOUT = DEFAULT_READ_TIMEOUT;
        public int maxConnections = DEFAULT_MAX_CONNECTIONS;

        public ImageCacheParams cacheParams;

        public ImageLoadConfig(Context context) {
            cacheParams = new ImageCacheParams(context);
        }

        public ImageLoadConfig(ImageCacheParams params) {
            cacheParams = params;
        }
    }

}
