package com.ichinaski.imageloader.cache;

import android.app.Activity;
import android.app.ActivityManager;
import android.content.Context;
import android.support.v4.util.LruCache;

import java.io.File;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import com.ichinaski.imageloader.BitmapSample;
import com.ichinaski.imageloader.ImageLoadUtils;
import com.ichinaski.imageloader.ImageSize;
import com.ichinaski.imageloader.util.IOUtils;
import com.ichinaski.imageloader.util.PlatformUtils;

import static com.ichinaski.imageloader.ImageLoader.TAG;
import static com.ichinaski.imageloader.util.LogUtils.LOGD;
import static com.ichinaski.imageloader.util.LogUtils.LOGE;
import static com.ichinaski.imageloader.util.LogUtils.LOGV;

public class ImageCache {

    private LruCache<String, BitmapSample> mMemoryCache;
    private DiskLruCache mDiskCache;
    private ImageCacheParams mCacheParams;
    private Context mContext; 
    private Object mDiskCacheLock = new Object();

    /**
     * Creating a new ImageCache object using the specified parameters.
     * 
     * @param context The context to use
     * @param cacheParams The cache parameters to use to initialize the cache
     */
    public ImageCache(Context context, ImageCacheParams cacheParams) {
        init(context, cacheParams);
    }

    /**
     * Creating a new ImageCache object using the default parameters.
     * 
     * @param context The context to use
     * @param uniqueName A unique name that will be appended to the cache
     *            directory
     */
    public ImageCache(Context context) {
        init(context, new ImageCacheParams(context));
    }

    /**
     * Initialize the cache, providing all parameters.
     * 
     * @param context The context to use
     * @param cacheParams The cache parameters to initialize the cache
     */
    private void init(Context context, ImageCacheParams cacheParams) {
        mContext = context;
        mCacheParams = cacheParams;

        // Initialize disk cache
        initDiskCache();
        
        // Set up memory cache
        initMemoryCache();
    }
    
    private void initDiskCache () {
        synchronized (mDiskCacheLock) {
            final File diskCacheDir = PlatformUtils.getDiskCacheDir(mContext, mCacheParams.uniqueName);

            if (mCacheParams.diskCacheEnabled) {
                if (!diskCacheDir.exists()) {
                    diskCacheDir.mkdir();
                }

                if (PlatformUtils.getUsableSpace(diskCacheDir) > mCacheParams.diskCacheSize) {
                    try {
                        mDiskCache = DiskLruCache.open(diskCacheDir, 1, 1, mCacheParams.diskCacheSize);
                        LOGD(TAG, "Disk Cache created");
                    } catch (final IOException e) {
                        LOGE(TAG, "DiskCache opening Error. - " + e);
                    }
                }
            }
            mDiskCacheLock.notifyAll();
        }
    }
    
    private void initMemoryCache () {
        if (mCacheParams.memoryCacheEnabled) {
            mMemoryCache = new BitmapLruCache(mCacheParams.memCacheSize);
            LOGD(TAG, "Memory Cache created");
        }
    }

    /**
     * Set whether file cache should be used.
     * If not, delete stored bitmaps.
     * 
     * @param useFileCache
     */
    public void setUseFileCache(boolean useFileCache) {
        if (mCacheParams != null) {
            mCacheParams.diskCacheEnabled = useFileCache;
        }
        
        if (!mCacheParams.diskCacheEnabled) {
            clearDiskCache();
        }
    }
    
    public void copyFileToDiskCache(String name, File file) {
        synchronized (mDiskCacheLock) {
            if (mDiskCache != null) {
                final String key = ImageLoadUtils.getImageId(name);
                InputStream is = null;
                OutputStream os = null;
                try {
                    final DiskLruCache.Editor editor = mDiskCache.edit(key);
                    if (editor != null) {
                        is = new FileInputStream(file);
                        os = editor.newOutputStream(ImageCacheParams.DISK_CACHE_INDEX);
                        IOUtils.copyStream(is, os);
                        editor.commit();
                    }
                } catch (final IOException e) {
                    LOGE(TAG, "addFileToDiskCache - " + e);
                } finally {
                    IOUtils.closeSilently(is);
                    IOUtils.closeSilently(os);
                }
            }
        }
    }

    /**
     * Add Bitmap to memory cache. This should receive the resized bitmap
     * 
     * @param data
     * @param bitmap
     */
    public synchronized void addBitmapToMemoryCache(String data, BitmapSample bitmapSample) {
        if (mMemoryCache != null && mMemoryCache.get(data) == null) {
            mMemoryCache.put(data, bitmapSample);
        }
    }

    /**
     * Get from memory cache.
     * 
     * @param data Unique identifier for which item to get
     * @return The bitmap if found in cache, null otherwise
     */
    public BitmapSample getBitmapFromMemCache(String url) {
        if (mMemoryCache != null) {
            return mMemoryCache.get(url);
        }
        return null;
    }

    /**
     * Get bitmap from disk cache.
     * 
     * @param name The name of the image to read
     * @param size The requested ImageSize of the resulting bitmap
     * @return The bitmap if found in cache, null otherwise
     */
    public BitmapSample getBitmapFromDiskCache(String name, ImageSize size) {
        final String key = ImageLoadUtils.getImageId(name);
        BitmapSample bitmapSample = null;
        FileInputStream fis = null;

        synchronized (mDiskCacheLock) {
            if (mDiskCache != null) {
                try {
                    final DiskLruCache.Snapshot snapshot = mDiskCache.get(key);
                    if (snapshot != null) {
                        LOGV(TAG, "Disk cache hit");
                        fis = (FileInputStream)snapshot.getInputStream(ImageCacheParams.DISK_CACHE_INDEX);
                        FileDescriptor descriptor = fis.getFD();
                        if (descriptor != null) {
                            bitmapSample = ImageLoadUtils.decodeSampledBitmap(descriptor, size);
                        } 
                    }
                } catch (final IOException e) {
                    LOGE(TAG, "getBitmapFromDiskCache - " + e);
                } finally {
                    IOUtils.closeSilently(fis);
                }
            }
        }
        return bitmapSample;
    }

    /**
     * Force buffered operations to the Disk Cache
     */
    public void close() {
        if (mDiskCache != null) {
            try {
                if (!mDiskCache.isClosed()) {
                    mDiskCache.flush();
                }
            } catch (IOException ignored) {}
        }
    }

    /**
     * Clear memory cache.
     */
    public void clearMemoryCache() {
        if (mMemoryCache != null) {
            mMemoryCache.evictAll();
        }
    }

    /**
     * Deletes disk cache.
     */
    public void clearDiskCache() {
        try {
            if (mDiskCache != null) {
                mDiskCache.delete();
                LOGD(TAG, "Disk Cache deleted");
            }
        } catch (IOException e) {
            LOGE(TAG, "clearDiskCache() - " + e);
        }
        
        // If necessary, restart cache
        initDiskCache();
    }

    /**
     * Clear both memory and disk cache.
     */
    public void clearCaches() {
        clearMemoryCache();
        clearDiskCache();
    }

    /**
     * A holder class that contains cache parameters.
     */
    public static class ImageCacheParams {
        // Default name in disk cache
        private static final String DEFAULT_CACHE_NAME = "imagesCache";
    
        // Default disk cache size
        private static final int DEFAULT_DISK_CACHE_SIZE = 1024 * 1024 * 10; // 10MB
        static final int DISK_CACHE_INDEX = 0;
    
        public String uniqueName = DEFAULT_CACHE_NAME;
        public long diskCacheSize = DEFAULT_DISK_CACHE_SIZE;
        public boolean memoryCacheEnabled = true;
        public boolean diskCacheEnabled = true;
        public int memCacheSize;

        public ImageCacheParams(Context context) {
            initSize(context);
        }
        
        private void initSize(Context context) {
            final ActivityManager activityManager = (ActivityManager)context
                    .getSystemService(Activity.ACTIVITY_SERVICE);
            int memoryClass = activityManager.getMemoryClass();
            
            // Size = 1/8 of available memory
            memCacheSize = memoryClass / 8 * 1024 * 1024;
        }
    }

    /**
     * Clear specified Bitmap from memory cache. Used when the Bitmap has been
     * scaled down and we need a bigger resolution. This allows the ImageLoader
     * to request it again (from disk/web) and use in its original size.
     * 
     * @param key
     */
    public void clearBitmap(String key) {
        if (mMemoryCache != null) {
            mMemoryCache.remove(key);
        }
    }

}
